open Base
open Ctypes

include Crypto_intf
include Crypto_intf.Exceptions
include Crypto_intf.Types

module type Stringable_carray = sig
  include Transmittable.S
  val of_char_carray: char CArray.t -> t
  val to_char_carray: t -> char CArray.t
end

module Stringable_carray: Stringable_carray = struct
  type t = char CArray.t

  let equal a b =
    Int.equal (CArray.length a) (CArray.length b) && begin
      let elts_equal = ref true in
      for i = 0 to CArray.length a - 1 do
        if !elts_equal
        then elts_equal := Char.equal (CArray.get a i) (CArray.get b i)
        else ()
      done;
      !elts_equal
    end

  let of_char_carray arr =
    arr

  let to_char_carray key =
    key

  let of_string str =
    CArray.of_list char (String.to_list str)

  let to_string key =
    String.of_char_list (CArray.to_list key)

  let of_base64_exn str =
    match Base64.decode ~alphabet:Base64.uri_safe_alphabet ~pad:false str with
    | Result.Ok result -> of_string result
    | Result.Error _ -> raise (Key_decoding_error str)

  let to_base64 key =
    key |> to_string |> Base64.encode_string ~alphabet:Base64.uri_safe_alphabet ~pad:false
end

module Secret_key: Stringable_carray = Stringable_carray
module Public_key: Stringable_carray = Stringable_carray
module Nonce: Stringable_carray = Stringable_carray
module Plain_text: Stringable_carray = Stringable_carray
module Cipher_text: Stringable_carray = Stringable_carray
module Signature: Stringable_carray = Stringable_carray

module Sodium = Sodium.Bindings(Sodium_generated)

let generate_keypair ~purpose =
  let generate_encryption_pair () =
    let pk_size = Sodium.crypto_box_publickeybytes () in
    let sk_size = Sodium.crypto_box_secretkeybytes () in
    let pk = CArray.make char pk_size in
    let sk = CArray.make char sk_size in
    Sodium.crypto_box_keypair (CArray.start pk) (CArray.start sk);
    (Public_key.of_char_carray pk, Secret_key.of_char_carray sk) in
  let generate_signing_pair () =
    let pk_size = Sodium.crypto_sign_publickeybytes () in
    let sk_size = Sodium.crypto_sign_secretkeybytes () in
    let pk = CArray.make char pk_size in
    let sk = CArray.make char sk_size in
    Sodium.crypto_sign_keypair (CArray.start pk) (CArray.start sk);
    (Public_key.of_char_carray pk, Secret_key.of_char_carray sk) in
  match purpose with
  | Encryption -> generate_encryption_pair ()
  | Signing -> generate_signing_pair ()

let generate_seed_keypair ~seed:seed_str =
  let pk_size = Sodium.crypto_box_publickeybytes () in
  let sk_size = Sodium.crypto_box_secretkeybytes () in
  let seed_size = Sodium.crypto_box_seedbytes () in
  let pk = CArray.make char pk_size in
  let sk = CArray.make char sk_size in
  let seed = CArray.make char seed_size ~initial:'\x00' in
  String.to_list seed_str
  |> List.iteri ~f:begin fun i c ->
    if i < seed_size then CArray.set seed i c;
  end;
  Sodium.crypto_box_seed_keypair (CArray.start pk) (CArray.start sk) (CArray.start seed);
  (Public_key.of_char_carray pk, Secret_key.of_char_carray sk)

let generate_nonce () =
  let nonce_size = Sodium.crypto_box_noncebytes () in
  let nonce = CArray.make char nonce_size in
  Sodium.randombytes_buf (CArray.start nonce) nonce_size;
  Nonce.of_char_carray nonce

let encrypt plaintext pk sk nonce =
  let plaintext_array = Plain_text.to_char_carray plaintext in
  let plaintext_size = CArray.length plaintext_array in
  if plaintext_size <= 0 then raise Plaintext_is_empty;
  let ciphertext_size = plaintext_size + Sodium.crypto_box_macbytes () in
  let ciphertext = CArray.make char ciphertext_size in
  let result = Sodium.crypto_box_easy
      (CArray.start ciphertext)
      (CArray.start plaintext_array)
      plaintext_size
      (CArray.start (Nonce.to_char_carray nonce))
      (CArray.start (Public_key.to_char_carray pk))
      (CArray.start (Secret_key.to_char_carray sk)) in
  if Int.equal result 0
  then Cipher_text.of_char_carray ciphertext
  else raise Encryption_error (* TODO: more fine-grained error reporting *)

let decrypt ciphertext pk sk nonce =
  let ciphertext_array = Cipher_text.to_char_carray ciphertext in
  let ciphertext_size = CArray.length ciphertext_array in
  let plaintext_size = ciphertext_size - Sodium.crypto_box_macbytes () in
  if plaintext_size <= 0 then raise Ciphertext_is_too_short;
  let plaintext = CArray.make char plaintext_size in
  let result = Sodium.crypto_box_open_easy
      (CArray.start plaintext)
      (CArray.start ciphertext_array)
      ciphertext_size
      (CArray.start (Nonce.to_char_carray nonce))
      (CArray.start (Public_key.to_char_carray pk))
      (CArray.start (Secret_key.to_char_carray sk)) in
  if Int.equal result 0
  then Plain_text.of_char_carray plaintext
  else raise Decryption_error (* TODO: more fine-grained error reporting *)

let sign msg sk =
  let sk_array = Secret_key.to_char_carray sk in
  let msg_array = CArray.of_list char (String.to_list msg) in
  let msg_size = CArray.length msg_array in
  let sig_size = Sodium.crypto_sign_bytes () in
  let sig_array = CArray.make char sig_size in
  let result = Sodium.crypto_sign_detached
      (CArray.start sig_array)
      (from_voidp ullong null)
      (CArray.start msg_array)
      msg_size
      (CArray.start sk_array) in
  if Int.equal result 0
  then Signature.of_char_carray sig_array
  else raise Signing_error

let verify_signature msg signt pk =
  let sig_array = Signature.to_char_carray signt in
  let pk_array = Public_key.to_char_carray pk in
  let msg_array = CArray.of_list char (String.to_list msg) in
  let msg_size = CArray.length msg_array in
  let result = Sodium.crypto_sign_verify_detached
      (CArray.start sig_array)
      (CArray.start msg_array)
      msg_size
      (CArray.start pk_array) in
  Int.equal result 0

let hash_password password =
  let password_array = CArray.of_list char (String.to_list password) in
  let password_size = CArray.length password_array in
  let hash_size = Sodium.crypto_pwhash_strbytes () in
  let hash_array = CArray.make char hash_size in
  let result = Sodium.crypto_pwhash_str
      (CArray.start hash_array)
      (CArray.start password_array)
      password_size
      (Sodium.crypto_pwhash_opslimit_sensitive ())
      (Sodium.crypto_pwhash_memlimit_sensitive ()) in
  if Int.equal result 0
  then String.of_char_list (CArray.to_list hash_array)
  else raise Hash_error

let verify_password ~password:password ~hash:hash =
  let password_array = CArray.of_list char (String.to_list password) in
  let password_size = CArray.length password_array in
  let hash_array = CArray.of_list char (String.to_list hash) in
  let result = Sodium.crypto_pwhash_str_verify
      (CArray.start hash_array)
      (CArray.start password_array)
      password_size in
  Int.equal result 0
