open Base

module Transmittable = struct
  module type S = sig
    type t
    include Stringable.S with type t := t
    include Equal.S with type t := t

    (** Encodes key data using web-safe Base64 encoding *)
    val to_base64: t -> string

    (** Decodes key data using web-safe Base64 decoding.
        Raises [Key_decoding_error] when input is not a valid Base64 string. *)
    val of_base64_exn: string -> t
  end
end

module Exceptions = struct
  exception Plaintext_is_empty
  exception Encryption_error
  exception Ciphertext_is_too_short
  exception Decryption_error
  exception Hash_error
  exception Signing_error
  exception Key_decoding_error of string
end

module Types = struct
  (** Purpose of the key: signing or encryption *)
  type key_purpose =
    | Encryption
    | Signing
end

module type Crypto = sig
  (** Interface for data transmitted over HTTP *)
  module Transmittable = Transmittable

  (** Container for secret key *)
  module Secret_key: Transmittable.S

  (** Container for public key *)
  module Public_key: Transmittable.S

  (** Container for nonce value *)
  module Nonce: Transmittable.S

  (** Container for plain text *)
  module Plain_text: Transmittable.S

  (** Container for cipher text *)
  module Cipher_text: Transmittable.S

  (** Container for message signature *)
  module Signature: Transmittable.S

  include module type of Exceptions

  include module type of Types

  (** Generates a pair of public and secret keys for encryption or signing *)
  val generate_keypair: purpose:key_purpose -> (Public_key.t * Secret_key.t)

  (** Generates a pair of public and secret encryption keys deterministically *)
  val generate_seed_keypair: seed:string -> (Public_key.t * Secret_key.t)

  (** Generates random nonce value *)
  val generate_nonce: unit -> Nonce.t

  (** Encrypts the plaintext using public key, secret key and nonce value.
      Raises [Plaintext_is_empty] when plaintext is empty.
      Raises [Encryption_error] whenever any other error occurs. *)
  val encrypt: Plain_text.t -> Public_key.t -> Secret_key.t -> Nonce.t -> Cipher_text.t

  (** Decrypts the ciphertext using public key, secret key and nonce value.
      Raises [Ciphertext_is_too_short] when ciphertext is empty or not long enough to decrypt.
      Raises [Decryption_error] whenever any other error occurs. *)
  val decrypt: Cipher_text.t -> Public_key.t -> Secret_key.t -> Nonce.t -> Plain_text.t

  (** Signs the message with the given key and returns the signature. *)
  val sign: string -> Secret_key.t -> Signature.t

  (** Verifies the message using given key and signature *)
  val verify_signature: string -> Signature.t -> Public_key.t -> bool

  (** Creates hash of the given password. Use it to store password in database. *)
  val hash_password: string -> string

  (** Compares password with its hash from [hash_password] *)
  val verify_password: password:string -> hash:string -> bool
end
