open Base
open Oksql

include User_intf
include User_intf.Types

type t = user [@@deriving equal]

module User_from_sql = struct
  type t = user

  let of_sql_row row =
    let open Sqlite3.Data in
    let open Crypto in
    match[@warning "-4"] row with
    | [| "id", TEXT id;
         "password_hash", TEXT password_hash;
         "public_key", TEXT pk;
         "secret_key", TEXT sk;
         "name", TEXT name;
      |] ->
      let public_key = Public_key.of_base64_exn pk in
      let secret_key = Secret_key.of_base64_exn sk in
      { id; password_hash; public_key; secret_key; name }
    | _ ->
      failwith "User.of_sql_row: invalid data"
end

let q_insert_user = {|
INSERT INTO users
(id, password_hash, public_key, secret_key, name)
VALUES
(:id, :password_hash, :public_key, :secret_key, :name);
|}

let q_find_all_users = {|
SELECT id, password_hash, public_key, secret_key, name
FROM users;
|}

let q_remove_user = {|
DELETE FROM users
WHERE id = :id;
|}

let q_find_by_id = {|
SELECT id, password_hash, public_key, secret_key, name
FROM users
WHERE id = :id;
|}

let q_change_password = {|
UPDATE users
SET password_hash = :password_hash
WHERE id = :id;
|}

let insert_user db user =
  let open Sqlite3.Data in
  let open Crypto in
  exec_only db q_insert_user ~params:[
    ":id", TEXT user.id;
    ":password_hash", TEXT user.password_hash;
    ":public_key", TEXT (Public_key.to_base64 user.public_key);
    ":secret_key", TEXT (Secret_key.to_base64 user.secret_key);
    ":name", TEXT user.name;
  ]

let find_all db =
  exec_fetch (module User_from_sql) db q_find_all_users

let create db ~id ~password ~name =
  let password_hash = Crypto.hash_password password in
  let (public_key, secret_key) = Crypto.generate_keypair ~purpose:Crypto.Encryption  in
  let user = { id; password_hash; public_key; secret_key; name } in
  insert_user db user;
  user

let remove db ~id =
  let open Sqlite3.Data in
  exec_only db q_remove_user ~params:[":id", TEXT id]

let find_by_id db ~id =
  let open Sqlite3.Data in
  let results =
    exec_fetch (module User_from_sql) db q_find_by_id ~params:[":id", TEXT id] in
  match results with
  | [| user |] -> Some user
  | [| |] -> None
  | _ -> failwith "User.find_by_id: more than one user is found"

let change_password db ~id ~new_password =
  let open Sqlite3.Data in
  let new_pw_hash = Crypto.hash_password new_password in
  exec_only db q_change_password ~params:[
    ":id", TEXT id;
    ":password_hash", TEXT new_pw_hash;
  ]
