open Alcotest
open Base
open Okauth.Crypto

let test_should_not_encrypt_empty_plaintext =
  test_case "encrypt should raise an exception when plaintext is empty" `Quick
    begin fun () ->
      check_raises "empty plaintext" Plaintext_is_empty begin fun () ->
        let (pk1, _) = generate_keypair ~purpose:Encryption in
        let (_, sk2) = generate_keypair ~purpose:Encryption in
        let nonce = generate_nonce () in
        ignore (encrypt (Plain_text.of_string "") pk1 sk2 nonce)
      end
    end

let test_roundtrip_encrypt_decrypt =
  let (pk1, sk1) = generate_keypair ~purpose:Encryption in
  let (pk2, sk2) = generate_keypair ~purpose:Encryption in
  QCheck.Test.make QCheck.printable_string
    ~name:"encrypted messages should be decrypted to the original one"
    begin fun msg ->
      QCheck.assume (Int.(0 <> String.length msg));
      let plaintext = Plain_text.of_string msg in
      let nonce = generate_nonce () in
      let ciphertext = encrypt plaintext pk1 sk2 nonce in
      let decrypted = decrypt ciphertext pk2 sk1 nonce in
      String.equal msg (Plain_text.to_string decrypted)
    end

let test_roundtrip_hash_verify =
  QCheck.Test.make QCheck.(pair string string)
    ~count:1 (* Because argon2 hashing algoritm takes _quite_ some time *)
    ~max_gen:10 ~max_fail:10
    ~name:"hash of the password should be verified by the same password only"
    begin function | (psw_a, psw_b) ->
      QCheck.assume (String.(psw_a <> psw_b));
      let hash_a = hash_password psw_a in
      (verify_password ~password:psw_a ~hash:hash_a) &&
      not (verify_password ~password:psw_b ~hash:hash_a)
    end

let test_roundtrip_sign_verify =
  let (pk, sk) = generate_keypair ~purpose:Signing in
  QCheck.Test.make QCheck.printable_string
    ~name:"signed messages should be verified"
    begin fun msg ->
      verify_signature msg (sign msg sk) pk
    end

let test_seed_keypair_generates_the_same_key =
  QCheck.Test.make QCheck.printable_string
    ~name:"generate_seed_keypair should generate the same keys for the same seed"
    begin fun seed ->
      let (pk1, sk1) = generate_seed_keypair ~seed in
      let (pk2, sk2) = generate_seed_keypair ~seed in
      (Public_key.equal pk1 pk2) && (Secret_key.equal sk1 sk2)
    end

let cases = List.concat [[
    test_should_not_encrypt_empty_plaintext;
  ]; List.map [
    test_roundtrip_encrypt_decrypt;
    test_roundtrip_hash_verify;
    test_roundtrip_sign_verify;
    test_seed_keypair_generates_the_same_key;
  ] ~f:QCheck_alcotest.to_alcotest]
